# -*- coding: utf-8 -*-

from abc import ABC, abstractmethod
from typing import Any, Dict, Iterator, List

from core_mixins.utils import convert_data_type
from core_mixins.utils import remove_attributes
from core_mixins.utils import rename_attributes

from .base import IBaseETL


class IBaseEtlFromRecord(IBaseETL, ABC):
    """
    Base class for an ETL task that need to do ETLs processes over data (records,
    rows) retrieved from different sources like: file, sFTP server, SQS queues, APIs
    or another data source...
    """

    def __init__(
            self, name_mapper: Dict[str, str] = None, type_mapper: Dict[str, str] = None,
            max_per_batch: int = 1000, attrs_to_remove: List[str] = None, **kwargs):

        """
        :param name_mapper: It defines the old and new name for the attributes.
        :param type_mapper: It defines if a casting must be applied to an attribute.
        :param max_per_batch: Maximum number of elements to process per batch.
        :param attrs_to_remove: Attributes to remove from the records.
        """

        super(IBaseEtlFromRecord, self).__init__(**kwargs)

        self.name_mapper = name_mapper or {}
        self.attrs_to_remove = attrs_to_remove or []
        self.type_mapper = type_mapper or {}
        self.max_per_batch = max_per_batch

    def _execute(self, *args, **kwargs) -> int:
        """ Process (by applying transformations) each record from the source """

        batch_number = 1
        count = 0

        for batch in self.retrieve_records(**kwargs):
            self.info(f"Processing batch # {batch_number}...")
            records = []

            for record in batch:
                # Apply transformations required before the base ones...
                self.pre_transformations(record)

                remove_attributes(record, self.attrs_to_remove)
                rename_attributes(record, self.name_mapper)
                convert_data_type(record, self.type_mapper)

                # Apply transformations required after the base ones...
                self.post_transformations(record)

                records.append(record)
                count += 1

            if records:
                self.process_records(records)

            batch_number += 1

        return count

    @abstractmethod
    def retrieve_records(
            self, last_processed: Any = None, start: Any = None,
            end: Any = None, **kwargs) -> Iterator[List[Dict]]:

        """
        It retrieves records from sources. It's expected to return batches of records
        because in this way we can manage the amount of records to process in each iteration
        and do not exhaust the resources...

        :return: An iterator that contains a list of records.
        """

    def pre_transformations(self, record: Dict) -> None:
        """ Transformations applied before built-in ones... """

    def post_transformations(self, record: Dict) -> None:
        """ Transformations applied after built-in ones... """

    @abstractmethod
    def process_records(self, records: List[Dict], **kwargs):
        """
        It must implement the actions to do with the records after
        transformations like archive in S3, send to an sFTP server, send to
        an SQS queue or a Kinesis stream...
        """



# -*- coding: utf-8 -*-

from abc import ABC, abstractmethod
from typing import Any, Iterator

from .base import IBaseETL


class IBaseEtlFromFile(IBaseETL, ABC):
    """
    Base class for an ETL task that need to do something with a
    file retrieved from a source. Like copy the file
    from sFTP to S3...
    """

    def _execute(self, *args, **kwargs) -> int:
        """ Process each file from paths """

        processed_files = 0

        for path in self.get_paths():
            try:
                self.info(f"Processing file in path: {path}...")
                self.process_file(path, *args, **kwargs)
                self.on_success(path)

                self.info(f"Processed!")
                processed_files += 1
            
            except Exception as error:
                self.error(f"Error processing the file: {path}. Error: {error}")
                self.on_error(path)

        return processed_files

    @abstractmethod
    def get_paths(self, last_processed: Any = None, *args, **kwargs) -> Iterator[str]:
        """
        It retrieves the file(s) from the source and return an iterator
        with the paths for the files...

        :return: An iterator that contains the file paths.
        """

    def process_file(self, path: str, *args, **kwargs):
        """ Doing something with the file """

    def on_success(self, path: str, **kwargs):
        """ Do something after the file is processed successfully, like archiving... """

    def on_error(self, path: str, **kwargs):
        """ Do something after an error is raised while processing the file... """



# -*- coding: utf-8 -*-


from abc import ABC
from typing import List, Dict, Type, Self

from core_mixins.exceptions import get_exception_data
from core_mixins.interfaces.task import ITask
from core_mixins.interfaces.task import TaskStatus


class IBaseETL(ITask, ABC):
    """
    Base class for an ETL task. A task defines the operations
    over a service or a platform from where you need to
    extract, transform and/or store information...

    This class defines some commons (maybe useful) attributes/methods for future
    implementations of the ETL tasks.
    """

    _impls: Dict[str, Type[Self]] = {}

    def __init__(self, time_zone: str = None, temp_folder: str = "/tmp", **kwargs):
        """
        :param time_zone: The time zone to use in the ETL for date/datetime processing.
        :param temp_folder: For task that need to store local files temporarily.
        """

        super(IBaseETL, self).__init__(**kwargs)

        self.status = TaskStatus.CREATED
        self.temp_folder = temp_folder
        self.time_zone = time_zone

    def execute(self, *args, **kwargs) -> TaskStatus:
        super(IBaseETL, self).execute()
        self.status = TaskStatus.EXECUTING

        try:
            self.info("Executing pre-processing actions...")
            self.pre_processing()

            self.info("Executing main flow...")
            res = self._execute()

            self.info("Executing post-processing actions...")
            self.post_processing()

        except Exception as error:
            type_, message, traceback = get_exception_data()
            self.error(f"{type_} -- {error}")
            self.error(traceback)

            self.save_logs(
                TaskStatus.SUCCESS,
                error_type=type_,
                error_message=message,
                error_traceback=traceback
            )

            return TaskStatus.ERROR

        else:
            self.save_logs(TaskStatus.SUCCESS, processed_elements=res)
            self.info(f"Were processed {res} elements.")
            self.info(f"Finished -> {self.name}!")
            return TaskStatus.SUCCESS

        finally:
            self.clean_resources()

    def pre_processing(self, *args, **kwargs):
        """ Pre-processing actions... """

    def _execute(self, *args, **kwargs) -> int:
        """
        Generic implementation for all the extract, transform
        and load processes...

        :return: It returns the number of elements were processed.
        """

    def post_processing(self, *args, **kwargs):
        """ Post-processing actions... """

    def save_logs(
            self, execution_status: TaskStatus, processed_elements: int = 0, error_type: str = "",
            error_message: str = "", error_traceback: List[Dict] = None, **kwargs):

        """
        Save executions logs/metadata if required...

        :param execution_status: The status of the execution.
        :param processed_elements: The number of processed elements.
        :param error_type: The error type if exists.
        :param error_message: The error message if exists.
        :param error_traceback: The error traceback if exists.
        """

    def clean_resources(self, *args, **kwargs) -> None:
        """ In case you need to close/remove some resources """
